@@ -70,8 +70,8 @@ ioctl_tree_new_from_bin(unsigned long id, const void *data, int ret)
 ioctl_tree *
 ioctl_tree_new_from_text(const char *line)
 {
-    static char lead_ws[1000];
-    static char ioctl_name[100];
+    static char lead_ws[1001];
+    static char ioctl_name[101];
     int ret, offset;
     unsigned long id;
     const ioctl_type *type;
@@ -202,6 +202,10 @@ ioctl_tree_read(FILE * f)
 	if (line[0] == '\n' || line[0] == '#')
 	    continue;
 
+        /* skip lines with metadata */
+        if (line[0] == '@')
+            continue;
+
 	node = ioctl_tree_new_from_text(line);
 	if (node == NULL) {
 	    DBG("ioctl_tree_read: failure to parse line: %s", line);
@@ -536,6 +540,75 @@ ioctl_simplestruct_in_execute(const ioctl_tree * node, unsigned long id, void *a
     return 0;
 }
 
+/***********************************
+ *
+ * ioctls with dynamic length struct data, but no pointers to substructs
+ *
+ ***********************************/
+
+static void
+ioctl_varlenstruct_init_from_bin(ioctl_tree * node, const void *data)
+{
+    size_t size = node->type->get_data_size(node->id, data);
+    DBG("ioctl_varlenstruct_init_from_bin: %s(%lX): size is %lu bytes\n", node->type->name, node->id, size);
+    node->data = malloc(size);
+    memcpy(node->data, data, size);
+}
+
+static int
+ioctl_varlenstruct_init_from_text(ioctl_tree * node, const char *data)
+{
+    size_t data_len = strlen(data) / 2;
+
+    node->data = malloc(data_len);
+
+    if (!read_hex(data, node->data, data_len)) {
+	fprintf(stderr, "ioctl_varlenstruct_init_from_text: failed to parse '%s'\n", data);
+	free(node->data);
+	return FALSE;
+    }
+
+    /* verify that the text data size actually matches get_data_size() */
+    size_t size = node->type->get_data_size(node->id, node->data);
+
+    if (size != data_len) {
+	fprintf(stderr, "ioctl_varlenstruct_init_from_text: ioctl %lX: expected data length %zu, but got %zu bytes from text data\n",
+		node->id, size, data_len);
+	free(node->data);
+	return FALSE;
+    }
+
+    return TRUE;
+}
+
+static void
+ioctl_varlenstruct_write(const ioctl_tree * node, FILE * f)
+{
+    assert(node->data != NULL);
+    write_hex(f, node->data, node->type->get_data_size(node->id, node->data));
+}
+
+static int
+ioctl_varlenstruct_equal(const ioctl_tree * n1, const ioctl_tree * n2)
+{
+    size_t size1 = n1->type->get_data_size(n1->id, n1->data);
+    size_t size2 = n2->type->get_data_size(n2->id, n2->data);
+    return n1->id == n2->id && size1 == size2 && memcmp(n1->data, n2->data, size1) == 0;
+}
+
+static int
+ioctl_varlenstruct_in_execute(const ioctl_tree * node, unsigned long id, void *arg, int *ret)
+{
+    if (id == node->id) {
+	size_t size = node->type->get_data_size(id, node->data);
+	memcpy(arg, node->data, size);
+	*ret = node->ret;
+	return 1;
+    }
+
+    return 0;
+}
+
 /***********************************
  *
  * USBDEVFS_REAPURB
@@ -752,31 +825,45 @@ ioctl_insertion_parent_stateless(ioctl_tree * tree, ioctl_tree * node)
     return tree;
 }
 
+
 /***********************************
  *
  * Known ioctls
  *
  ***********************************/
 
+/* ioctl which does not need to consider its data argument and has no state */
 #define I_NOSTATE(name, execute_result) \
     {name, -1, 0, #name, NULL, NULL, NULL, NULL, NULL, ioctl_execute_ ## execute_result, NULL}
 
+/* input structs with fixed length with explicit size */
 #define I_NAMED_SIZED_SIMPLE_STRUCT_IN(name, namestr, size, nr_range, insertion_parent_fn) \
     {name, size, nr_range, namestr,                                            \
      ioctl_simplestruct_init_from_bin, ioctl_simplestruct_init_from_text,      \
      ioctl_simplestruct_free_data,                                             \
      ioctl_simplestruct_write, ioctl_simplestruct_equal,                       \
-     ioctl_simplestruct_in_execute, insertion_parent_fn}
-
-#define I_NAMED_SIMPLE_STRUCT_IN(name, namestr, nr_range, insertion_parent_fn) \
-    I_NAMED_SIZED_SIMPLE_STRUCT_IN(name, namestr, -1, nr_range, insertion_parent_fn)
+     ioctl_simplestruct_in_execute, insertion_parent_fn, NULL}
 
 #define I_SIZED_SIMPLE_STRUCT_IN(name, size, nr_range, insertion_parent_fn) \
     I_NAMED_SIZED_SIMPLE_STRUCT_IN(name, #name, size, nr_range, insertion_parent_fn)
 
+/* input structs with fixed length with size encoded in ioctl code */
+#define I_NAMED_SIMPLE_STRUCT_IN(name, namestr, nr_range, insertion_parent_fn) \
+    I_NAMED_SIZED_SIMPLE_STRUCT_IN(name, namestr, -1, nr_range, insertion_parent_fn)
+
 #define I_SIMPLE_STRUCT_IN(name, nr_range, insertion_parent_fn) \
     I_NAMED_SIZED_SIMPLE_STRUCT_IN(name, #name, -1, nr_range, insertion_parent_fn)
 
+/* input structs with a variable length (but no pointers to substructures) */
+#define I_VARLEN_STRUCT_IN(name, insertion_parent_fn, data_size_fn) \
+    {name, -1, 0, #name,                                                       \
+     ioctl_varlenstruct_init_from_bin, ioctl_varlenstruct_init_from_text,      \
+     ioctl_simplestruct_free_data,                                             \
+     ioctl_varlenstruct_write, ioctl_varlenstruct_equal,                       \
+     ioctl_varlenstruct_in_execute, insertion_parent_fn, data_size_fn}
+
+/* data with custom handlers; necessary for structs with pointers to nested
+ * structs, or keeping stateful handlers */
 #define I_CUSTOM(name, nr_range, fn_prefix)                     \
     {name, -1, nr_range, #name,                                 \
      fn_prefix ## _init_from_bin, fn_prefix ## _init_from_text, \
