@@ -26,6 +26,7 @@
 #include <sys/socket.h>
 #include <arpa/inet.h>
 #include <linux/un.h>
+#include <unistd.h>
 
 #include <libudev.h>
 
@@ -94,16 +95,19 @@ static void
 sendmsg_all(uevent_sender * sender, struct msghdr *msg)
 {
     glob_t gl;
+    int res;
 
     /* find current listeners */
-    if (glob(sender->socket_glob, GLOB_NOSORT, NULL, &gl) == 0) {
+    res = glob(sender->socket_glob, GLOB_NOSORT, NULL, &gl);
+    if (res == 0) {
 	size_t i;
 	for (i = 0; i < gl.gl_pathc; ++i)
 	    sendmsg_one(sender, msg, gl.gl_pathv[i]);
     } else {
 	/* ensure that we only fail due to that, not due to bad globs */
-	if (errno != GLOB_NOMATCH) {
-	    perror("sendmsg_all: cannot run glob");
+	if (res != GLOB_NOMATCH) {
+            fprintf(stderr, "ERROR: sendmsg_all: %s glob failed with %i\n",
+                    sender->socket_glob, res);
 	    abort();
 	}
     }
@@ -218,6 +222,12 @@ uevent_sender_send(uevent_sender * sender, const char *devpath, const char *acti
     count += strlen(props + count) + 1;
     strcpy(props + count, "SUBSYSTEM=");
     strcat(props + count, subsystem);
+    count += strlen(props + count) + 1;
+    if (udev_device_get_devnode(device)) {
+        strcpy(props + count, "DEVNAME=");
+        strcat(props + count, udev_device_get_devnode(device));
+        count += strlen(props + count) + 1;
+    }
 
     /* add versioned header */
     memset(&nlh, 0x00, sizeof(struct udev_monitor_netlink_header));
@@ -239,9 +249,9 @@ uevent_sender_send(uevent_sender * sender, const char *devpath, const char *acti
 
     /* add properties list */
     nlh.properties_off = iov[0].iov_len;
-    nlh.properties_len = sizeof(props);
+    nlh.properties_len = count;
     iov[1].iov_base = props;
-    iov[1].iov_len = sizeof(props);
+    iov[1].iov_len = count;
 
     /* send message */
     memset(&smsg, 0x00, sizeof(struct msghdr));
