@@ -28,6 +28,7 @@
 #include <sys/socket.h>
 #include <sys/un.h>
 #include <linux/usbdevice_fs.h>
+#include <linux/input.h>
 
 #include <libudev.h>
 #include <gudev/gudev.h>
@@ -135,11 +136,31 @@ t_testbed_add_devicev(UMockdevTestbedFixture * fixture, gconstpointer data)
     gchar *syspath;
     gchar *attributes[] = { "idVendor", "0815", "idProduct", "AFFE", NULL };
     gchar *properties[] = { "ID_INPUT", "1", "ID_INPUT_KEYBOARD", "1", NULL };
+    struct udev *udev;
+    struct udev_monitor *udev_mon;
+    struct udev_device *device;
+
+    /* Set up Udev monitor to check for added event */
+    udev = udev_new();
+    g_assert(udev != NULL);
+    udev_mon = udev_monitor_new_from_netlink(udev, "udev");
+    g_assert(udev_mon != NULL);
+    g_assert_cmpint(udev_monitor_get_fd(udev_mon), >, 0);
+    g_assert_cmpint(udev_monitor_enable_receiving(udev_mon), ==, 0);
 
     syspath = umockdev_testbed_add_devicev(fixture->testbed, "usb", "extkeyboard1", NULL, attributes, properties);
     g_assert_cmpstr(syspath, ==, "/sys/devices/extkeyboard1");
 
     _t_testbed_check_extkeyboard1(syspath);
+
+    device = udev_monitor_receive_device(udev_mon);
+    g_assert(device != NULL);
+    g_assert_cmpstr(udev_device_get_syspath(device), ==, syspath);
+    g_assert_cmpstr(udev_device_get_action(device), ==, "add");
+    udev_device_unref(device);
+    udev_monitor_unref(udev_mon);
+    udev_unref(udev);
+
     g_free(syspath);
 }
 
@@ -148,6 +169,17 @@ static void
 t_testbed_add_device(UMockdevTestbedFixture * fixture, gconstpointer data)
 {
     gchar *syspath;
+    struct udev *udev;
+    struct udev_monitor *udev_mon;
+    struct udev_device *device;
+
+    /* Set up Udev monitor to check for added event */
+    udev = udev_new();
+    g_assert(udev != NULL);
+    udev_mon = udev_monitor_new_from_netlink(udev, "udev");
+    g_assert(udev_mon != NULL);
+    g_assert_cmpint(udev_monitor_get_fd(udev_mon), >, 0);
+    g_assert_cmpint(udev_monitor_enable_receiving(udev_mon), ==, 0);
 
     syspath = umockdev_testbed_add_device(fixture->testbed, "usb", "extkeyboard1", NULL,
 					  /* attributes */
@@ -158,6 +190,15 @@ t_testbed_add_device(UMockdevTestbedFixture * fixture, gconstpointer data)
     g_assert(g_str_has_suffix(syspath, "/sys/devices/extkeyboard1"));
 
     _t_testbed_check_extkeyboard1(syspath);
+
+    device = udev_monitor_receive_device(udev_mon);
+    g_assert(device != NULL);
+    g_assert_cmpstr(udev_device_get_syspath(device), ==, syspath);
+    g_assert_cmpstr(udev_device_get_action(device), ==, "add");
+    udev_device_unref(device);
+    udev_monitor_unref(udev_mon);
+    udev_unref(udev);
+
     g_free(syspath);
 }
 
@@ -580,6 +621,18 @@ t_testbed_uevent_gudev(UMockdevTestbedFixture * fixture, gconstpointer data)
     g_free(syspath);
 }
 
+static void
+t_testbed_uevent_no_listener(UMockdevTestbedFixture * fixture, gconstpointer data)
+{
+    gchar *syspath;
+
+    syspath = umockdev_testbed_add_device(fixture->testbed, "pci", "mydev", NULL, NULL, NULL);
+    g_assert(syspath);
+
+    /* generate event */
+    umockdev_testbed_uevent(fixture->testbed, syspath, "add");
+}
+
 static void
 t_testbed_uevent_error(UMockdevTestbedFixture * fixture, gconstpointer data)
 {
@@ -1037,6 +1090,78 @@ t_testbed_dev_query_gudev(UMockdevTestbedFixture * fixture, gconstpointer data)
   g_assert_cmpint(errno, ==, EAGAIN);           \
   errno = 0;
 
+static void
+t_testbed_script_replay_evdev_event_framing(UMockdevTestbedFixture * fixture, gconstpointer data)
+{
+  gboolean success;
+  GError *error = NULL;
+  char *tmppath;
+  int fd;
+  char buf[1024];
+
+  struct timeval dummy = {0};
+  /* Simple evdev stream - x coordinate followed by SYN, times 2 */
+  struct input_event dummy_events[] = {
+      {dummy, 0003, 0000, 2534},
+      {dummy, 0000, 0000, 0000},
+      {dummy, 0003, 0000, 2200},
+      {dummy, 0000, 0000, 0000}
+  };
+
+  umockdev_testbed_add_from_string(fixture->testbed,
+          "P: /devices/event1\nN: input/event1\n"
+          "E: DEVNAME=/dev/input/event1\nE: SUBSYSTEM=input\n", &error);
+  g_assert_no_error(error);
+
+  /* write script into temporary file */
+  fd = g_file_open_tmp("test_script_simple.XXXXXX", &tmppath, &error);
+  g_assert_no_error(error);
+  g_assert_cmpint(write(fd, "r 0 ", strlen("r 0 ")), ==, strlen("r 0 "));
+  for (int i = 0; i < sizeof dummy_events; ++i) {
+    char cur = ((char *)dummy_events)[i];
+    if (cur < 32) {
+      g_assert_cmpint(write(fd, "^", 1), ==, 1);
+      cur += 64;
+    } else if (cur == '^') {
+      g_assert_cmpint(write(fd, "^", 1), ==, 1);
+      cur = '`';
+    }
+    g_assert_cmpint(write(fd, &cur, 1), ==, 1);
+  }
+  close(fd);
+
+  /* load it */
+  success = umockdev_testbed_load_script(fixture->testbed, "/dev/input/event1", tmppath, &error);
+  g_assert_no_error(error);
+  g_assert(success);
+  g_unlink (tmppath);
+
+  /* start communication */
+  fd = g_open("/dev/input/event1", O_RDWR | O_NONBLOCK, 0);
+  g_assert_cmpint(fd, >=, 0);
+
+  /* Evdev guarantees that we only read whole events
+   * We should do the same
+   */
+  ssize_t bytes_read = 0;
+  while (bytes_read < sizeof dummy_events) {
+    ssize_t tmp = read(fd, buf, sizeof buf);
+    if (tmp < 0) {
+      g_assert_cmpint(errno, ==, EAGAIN);
+      errno = 0;
+      continue;
+    }
+    bytes_read += tmp;
+    g_assert_cmpint(bytes_read % sizeof(struct input_event), ==, 0);
+  }
+  g_assert_cmpint(memcmp(buf, dummy_events, sizeof dummy_events), ==, 0);
+
+  /* end of script */
+  ASSERT_EOF;
+
+  close(fd);
+}
+
 static void
 t_testbed_script_replay_simple(UMockdevTestbedFixture * fixture, gconstpointer data)
 {
@@ -1102,11 +1227,11 @@ r 0 ^@^^^`^@a\n";
 
   /* now we should get the response after 10 ms */
   ASSERT_EOF;
-  usleep(15000);
+  usleep(20000);
   g_assert_cmpint(read(fd, buf, 11), ==, 11);
   g_assert(strncmp(buf, "response\t1\n", 11) == 0);
   g_assert_cmpint(errno, ==, 0);
-  usleep(2000);
+  usleep(5000);
   g_assert_cmpint(read(fd, buf, 15), ==, 15);
   g_assert(strncmp(buf, "hello world ☺\n", 15) == 0);
   g_assert_cmpint(errno, ==, 0);
@@ -1151,6 +1276,86 @@ r 0 ^@^^^`^@a\n";
   close(fd);
 }
 
+static void
+t_testbed_script_replay_default_device(UMockdevTestbedFixture * fixture, gconstpointer data)
+{
+  gboolean success;
+  GError *error = NULL;
+  char *tmppath;
+  int fd;
+  char buf[1024];
+
+  static const char* test_script = "d 0 /dev/greeter\n\
+r 0 OK\n";
+
+  umockdev_testbed_add_from_string(fixture->testbed,
+          "P: /devices/greeter\nN: greeter\n"
+          "E: DEVNAME=/dev/greeter\nE: SUBSYSTEM=tty\nA: dev=4:64\n", &error);
+  g_assert_no_error(error);
+
+  /* write script into temporary file */
+  fd = g_file_open_tmp("test_script_simple.XXXXXX", &tmppath, &error);
+  g_assert_no_error(error);
+  g_assert_cmpint(write(fd, test_script, strlen(test_script)), >, 10);
+  close(fd);
+
+  /* load it */
+  success = umockdev_testbed_load_script(fixture->testbed, NULL, tmppath, &error);
+  g_assert_no_error(error);
+  g_assert(success);
+  g_unlink (tmppath);
+
+  /* start communication */
+  fd = g_open("/dev/greeter", O_RDWR, 0);
+  g_assert_cmpint(fd, >=, 0);
+
+  /* check that we've got the right thing on the other end */
+  g_assert_cmpint(read(fd, buf, 2), ==, 2);
+  g_assert_cmpint(memcmp(buf, "OK", 2), ==, 0);
+
+  close(fd);
+}
+
+static void
+t_testbed_script_replay_override_default_device(UMockdevTestbedFixture * fixture, gconstpointer data)
+{
+  gboolean success;
+  GError *error = NULL;
+  char *tmppath;
+  int fd;
+  char buf[1024];
+
+  static const char* test_script = "d 0 /dev/bananas\n\
+r 0 OK\n";
+
+  umockdev_testbed_add_from_string(fixture->testbed,
+          "P: /devices/greeter\nN: greeter\n"
+          "E: DEVNAME=/dev/greeter\nE: SUBSYSTEM=tty\nA: dev=4:64\n", &error);
+  g_assert_no_error(error);
+
+  /* write script into temporary file */
+  fd = g_file_open_tmp("test_script_simple.XXXXXX", &tmppath, &error);
+  g_assert_no_error(error);
+  g_assert_cmpint(write(fd, test_script, strlen(test_script)), >, 10);
+  close(fd);
+
+  /* load it */
+  success = umockdev_testbed_load_script(fixture->testbed, "/dev/greeter", tmppath, &error);
+  g_assert_no_error(error);
+  g_assert(success);
+  g_unlink (tmppath);
+
+  /* start communication */
+  fd = g_open("/dev/greeter", O_RDWR, 0);
+  g_assert_cmpint(fd, >=, 0);
+
+  /* check that we've got the right thing on the other end */
+  g_assert_cmpint(read(fd, buf, 2), ==, 2);
+  g_assert_cmpint(memcmp(buf, "OK", 2), ==, 0);
+
+  close(fd);
+}
+
 static void
 t_testbed_script_replay_socket_stream(UMockdevTestbedFixture * fixture, gconstpointer data)
 {
@@ -1273,6 +1478,145 @@ r 0 OK\n";
   close(fd);
 }
 
+/* assert that difference between first and second is "ms" milliseconds, ±20 ms */
+static void
+assert_delta_t(const struct timeval * first, const struct timeval * second, int ms)
+{
+    int delta = (second->tv_sec - first->tv_sec) * 1000;
+    delta += (second->tv_usec - first->tv_usec) / 1000;
+    g_assert_cmpint(delta, >=, ms - 20);
+    g_assert_cmpint(delta, <=, ms + 20);
+}
+
+static void
+t_testbed_replay_evemu_events(UMockdevTestbedFixture * fixture, gconstpointer data)
+{
+  gboolean success;
+  GError *error = NULL;
+  int fd;
+  char *tmppath;
+  struct timeval tv_begin, tv_end;
+  struct input_event ev;
+  static const char* test_data = "E: 1234.500000 0000 0000 0\n"   /* SYN */
+                                 "E: 1234.650000 0002 0000 5  # comment\n"   /* REL X */
+                                 "# some comment\n"
+                                 /* slightly go back in time here */
+                                 "E: 1234.649000 0002 0001 -11	# comment #2\n" /* REL Y */
+                                 "E: 1235.200000 0001 0174 1\n";  /* KEY_ZOOM */
+
+  umockdev_testbed_add_from_string(fixture->testbed,
+          "P: /devices/event1\nN: input/event1\n"
+          "E: DEVNAME=/dev/input/event1\nE: SUBSYSTEM=input\n", &error);
+  g_assert_no_error(error);
+
+  /* write evemu events file */
+  fd = g_file_open_tmp("test_evemu.XXXXXX", &tmppath, &error);
+  g_assert_no_error(error);
+  g_assert_cmpint(write(fd, test_data, strlen(test_data)), ==, strlen(test_data));
+  close(fd);
+
+  /* load it */
+  success = umockdev_testbed_load_evemu_events(fixture->testbed, "/dev/input/event1", tmppath, &error);
+  g_assert_no_error(error);
+  g_assert(success);
+  g_unlink (tmppath);
+
+  /* start communication */
+  fd = g_open("/dev/input/event1", O_RDONLY, 0);
+  g_assert_cmpint(fd, >=, 0);
+
+  g_assert_cmpint(gettimeofday(&tv_begin, NULL), ==, 0);
+
+  /* read SYN event; that should happen immediately */
+  g_assert_cmpint(read(fd, &ev, sizeof(ev)), ==, sizeof(ev));
+  g_assert_cmpint(gettimeofday(&tv_end, NULL), ==, 0);
+  assert_delta_t(&tv_begin, &tv_end, 10);
+  tv_begin = tv_end;
+  g_assert_cmpint(ev.time.tv_sec, ==, 1234);
+  g_assert_cmpint(ev.time.tv_usec, ==, 500000);
+  g_assert_cmpint(ev.type, ==, 0);
+  g_assert_cmpint(ev.code, ==, 0);
+  g_assert_cmpint(ev.value, ==, 0);
+
+  /* read RELX event, should happen after 150 ms */
+  g_assert_cmpint(read(fd, &ev, sizeof(ev)), ==, sizeof(ev));
+  g_assert_cmpint(gettimeofday(&tv_end, NULL), ==, 0);
+  g_assert_cmpint(ev.time.tv_sec, ==, 1234);
+  g_assert_cmpint(ev.time.tv_usec, ==, 650000);
+  g_assert_cmpint(ev.type, ==, 2);
+  g_assert_cmpint(ev.code, ==, 0);
+  g_assert_cmpint(ev.value, ==, 5);
+  assert_delta_t(&tv_begin, &tv_end, 150);
+  tv_begin = tv_end;
+
+  /* read RELY event, should happen immediately after */
+  g_assert_cmpint(read(fd, &ev, sizeof(ev)), ==, sizeof(ev));
+  g_assert_cmpint(gettimeofday(&tv_end, NULL), ==, 0);
+  g_assert_cmpint(ev.time.tv_sec, ==, 1234);
+  g_assert_cmpint(ev.time.tv_usec, ==, 649000);
+  g_assert_cmpint(ev.type, ==, 2);
+  g_assert_cmpint(ev.code, ==, 1);
+  g_assert_cmpint(ev.value, ==, -11);
+  assert_delta_t(&tv_begin, &tv_end, 0);
+  tv_begin = tv_end;
+
+  /* read KEY_ZOOM event, should happen after 550 ms */
+  g_assert_cmpint(read(fd, &ev, sizeof(ev)), ==, sizeof(ev));
+  g_assert_cmpint(gettimeofday(&tv_end, NULL), ==, 0);
+  g_assert_cmpint(ev.time.tv_sec, ==, 1235);
+  g_assert_cmpint(ev.time.tv_usec, ==, 200000);
+  g_assert_cmpint(ev.type, ==, 1);
+  g_assert_cmpint(ev.code, ==, 0x174);
+  g_assert_cmpint(ev.value, ==, 1);
+  assert_delta_t(&tv_begin, &tv_end, 550);
+
+  close(fd);
+}
+
+static void
+t_testbed_replay_evemu_events_default_device(UMockdevTestbedFixture * fixture, gconstpointer data)
+{
+  gboolean success;
+  GError *error = NULL;
+  int fd;
+  char *tmppath;
+  struct input_event ev;
+  static const char* test_data = "# EVEMU 1.2\n"
+                                 "# device /dev/input/event1\n"
+                                 "# blabla\n"
+                                 "E: 1234.500000 0000 0000 0\n";  /* SYN */
+
+  umockdev_testbed_add_from_string(fixture->testbed,
+          "P: /devices/event1\nN: input/event1\n"
+          "E: DEVNAME=/dev/input/event1\nE: SUBSYSTEM=input\n", &error);
+  g_assert_no_error(error);
+
+  /* write evemu events file */
+  fd = g_file_open_tmp("test_evemu.XXXXXX", &tmppath, &error);
+  g_assert_no_error(error);
+  g_assert_cmpint(write(fd, test_data, strlen(test_data)), ==, strlen(test_data));
+  close(fd);
+
+  /* load it */
+  success = umockdev_testbed_load_evemu_events(fixture->testbed, NULL, tmppath, &error);
+  g_assert_no_error(error);
+  g_assert(success);
+  g_unlink (tmppath);
+
+  /* start communication */
+  fd = g_open("/dev/input/event1", O_RDONLY, 0);
+  g_assert_cmpint(fd, >=, 0);
+
+  /* read SYN event; that should happen immediately */
+  g_assert_cmpint(read(fd, &ev, sizeof(ev)), ==, sizeof(ev));
+  g_assert_cmpint(ev.time.tv_sec, ==, 1234);
+  g_assert_cmpint(ev.time.tv_usec, ==, 500000);
+  g_assert_cmpint(ev.type, ==, 0);
+  g_assert_cmpint(ev.code, ==, 0);
+  g_assert_cmpint(ev.value, ==, 0);
+
+  close(fd);
+}
 
 static void
 t_testbed_clear(UMockdevTestbedFixture * fixture, gconstpointer data)
@@ -1455,6 +1799,8 @@ main(int argc, char **argv)
 	       t_testbed_uevent_libudev, t_testbed_fixture_teardown);
     g_test_add("/umockdev-testbed/uevent/gudev", UMockdevTestbedFixture, NULL, t_testbed_fixture_setup,
 	       t_testbed_uevent_gudev, t_testbed_fixture_teardown);
+    g_test_add("/umockdev-testbed/uevent/no_listener", UMockdevTestbedFixture, NULL, t_testbed_fixture_setup,
+	       t_testbed_uevent_no_listener, t_testbed_fixture_teardown);
     g_test_add("/umockdev-testbed/uevent/error", UMockdevTestbedFixture, NULL, t_testbed_fixture_setup,
 	       t_testbed_uevent_error, t_testbed_fixture_teardown);
 
@@ -1475,10 +1821,20 @@ main(int argc, char **argv)
     /* tests for script replay */
     g_test_add("/umockdev-testbed/script_replay_simple", UMockdevTestbedFixture, NULL, t_testbed_fixture_setup,
 	       t_testbed_script_replay_simple, t_testbed_fixture_teardown);
+    g_test_add("/umockdev-testbed/script_replay_default_device", UMockdevTestbedFixture, NULL, t_testbed_fixture_setup,
+	       t_testbed_script_replay_default_device, t_testbed_fixture_teardown);
+    g_test_add("/umockdev-testbed/script_replay_override_default_device", UMockdevTestbedFixture, NULL, t_testbed_fixture_setup,
+	       t_testbed_script_replay_override_default_device, t_testbed_fixture_teardown);
+    g_test_add("/umockdev-testbed/script_replay_evdev_event_framing", UMockdevTestbedFixture, NULL, t_testbed_fixture_setup,
+               t_testbed_script_replay_evdev_event_framing, t_testbed_fixture_teardown);
     g_test_add("/umockdev-testbed/script_replay_socket_stream", UMockdevTestbedFixture, NULL, t_testbed_fixture_setup,
 	       t_testbed_script_replay_socket_stream, t_testbed_fixture_teardown);
     g_test_add("/umockdev-testbed/script_replay_fuzz", UMockdevTestbedFixture, NULL, t_testbed_fixture_setup,
 	       t_testbed_script_replay_fuzz, t_testbed_fixture_teardown);
+    g_test_add("/umockdev-testbed/replay_evemu_events", UMockdevTestbedFixture, NULL, t_testbed_fixture_setup,
+	       t_testbed_replay_evemu_events, t_testbed_fixture_teardown);
+    g_test_add("/umockdev-testbed/replay_evemu_events_default_device", UMockdevTestbedFixture, NULL, t_testbed_fixture_setup,
+	       t_testbed_replay_evemu_events_default_device, t_testbed_fixture_teardown);
 
     /* misc */
     g_test_add("/umockdev-testbed/clear", UMockdevTestbedFixture, NULL, t_testbed_fixture_setup,
